#include "DifferentialEvolution.h"

//Constructor
DifferentialEvolution::DifferentialEvolution(OptimizationProblemInterface* op, int g_max, int np, double Cr) : rng(time(0)), dist(0, 1.0) { /*Motor y tiempo son temporales hasta conocer los limites*/
	//Optimization problem
	this->op = op;
	//Optimization problem parameters
	this->d = op->D(); // Problem dimensionality (number of design variables)
	this->d_min = op->Dmin(); // Lower bounds of design variables
	this->d_max = op->Dmax(); // Upper bounds of design variables

	//Algorithm parameters
	this->np = np;//Population size
	this->g_max = g_max;//Number of generations
	this->Cr = Cr; //Crossover rate

	//Initialize population matrix
	this->x = vector<vector<double>>(np);
	for (int i = 0; i < np; i++) {
		this->x[i] = vector<double>(d);
	}
	//Initialize population fitness vector
	this->f_x = vector<double>(np);

	//Initialize population constraints vector
	this->g_x = vector<double>(np);
}

DifferentialEvolution::~DifferentialEvolution() {
}


vector<double> DifferentialEvolution::startAlgorithm() {
	//Generate a new population randomly
	initializePopulation();

	//Evaluate the population
	vector<double> result; // Auxiliar variable to store the evaluation result
	for (int i = 0; i < np; i++) {
		result = op->evaluate(x[i]); // Evaluate objective function and constraints
		f_x[i] = result[0]; // Get and store objective function value
		g_x[i] = result[1]; // Get and store constraints value
	}

	int r1 = 0, r2 = 0, r3 = 0; // Variables to store random individual indices required by DE/rand/1/bin
	int jrand = 0; // Variables to store a random variable index required by DE/rand/1/bin

	// Start evolutionary cycle
	for (int gen = 0; gen < g_max; gen++) {
		//Update scaling factor
		F = 0.3 + scaledRand()*(0.9 - 0.3);

		// For each individual in the population
		for (int i = 0; i < np; i++) {
			// Create a new individual using the information of three random parents r1, r2, and r3 such that i != r1 != r2 != r3
			do { // Find r1
				r1 = rng() % np;
			} while (r1 == i);

			do { // Find r2
				r2 = rng() % np;
			} while (r2 == r1 || r2 == i);

			do { // Find r3
				r3 = rng() % np;
			} while (r3 == r2 || r3 == r1 || r3 == i);

			jrand = rng() % d;  // Select a random variable index jrand for the bin crossover

			// Perform mutation and crossover to generate a new offspring individual u
			vector<double> u = vector<double>(d); // Offspring
			double f_u = 0; // Objective function value of the offspring
			double g_u = 0; // Constraints value of the offspring

			// For each design variable
			for (int j = 0; j < d; j++) { //bin
				if (scaledRand() < Cr || j == jrand) { // If this probability conditions are met
					// Generate a new design variable for the offspring (rand/1)
					u[j] = x[r1][j] + F * (x[r2][j] - x[r3][j]);

					// Repair the variable if it is out of the bounds
					if (u[j] < d_min[j] || u[j] > d_max[j]) {
						u[j] = d_min[j] + scaledRand() * (d_max[j] - d_min[j]); // Generate a new variable randomly within the bounds
					}
				}
				else { // Otherwise
					u[j] = x[i][j]; // Use the original design variable in the offspring
				}
			}

			// Evaluate offspring
			result = op->evaluate(u); // Evaluate objective function and constraints
			f_u = result[0]; // Get and store objective function value
			g_u = result[1]; // Get and store constraints value

			// Select the best between the original individual xi and the offspring u
			// If xi is more feasible than u
			if (g_x[i] < g_u) {
				// Xi is the best
				//nop
			}
			// If ui is more feasible than xi
			else if (g_x[i] > g_u) {
				// u is the best
				x[i] = u;
				f_x[i] = f_u;
				g_x[i] = g_u;
			}
			// If both are feasible
			else if (g_x[i] == 0 && g_u == 0) {
				// Check the objective function values
				// If xi has a better objective function value than u
				if (f_x[i] < f_u) {
					// Xi is the best
					//nop
				}
				// If u has a better objective function value than xi
				else {
					// u is the best
					x[i] = u;
					f_x[i] = f_u;
					g_x[i] = g_u;
				}
			}
			// If both are equally unfeasible
			else if (scaledRand() > 0.5) { // Select the best based on probability
				// Xi is the best
				//nop
			}
			else {
				// u is the best
				x[i] = u;
				f_x[i] = f_u;
				g_x[i] = g_u;
			}

		}
	}

	//Find the best individual in the final population
	int bestIndex = 0; // Index of the best individual
	// For each individual in population
	for (int i = 1; i < np; i++) {
		//Check if the currrent individual is the best based on the feasibility rules (Deb rules)
		if (g_x[i] < g_x[bestIndex]) {
			bestIndex = i;
		}
		else if (g_x[i] > g_x[bestIndex]) {
			//nop
		}
		else if (g_x[i] == 0 && g_x[bestIndex] == 0) {
			if (f_x[i] < f_x[bestIndex]) {
				bestIndex = i;
			}
			else {
				//nop
			}
		}
		else if (scaledRand() < 0.5) {
			bestIndex = i;
		}
		else {
			//nop
		}
	}

	return x[bestIndex];
}

//Generate a new population randomly
void DifferentialEvolution::initializePopulation() {
	for (int i = 0; i < np; i++) {
		for (int j = 0; j < d; j++) {
			x[i][j] = d_min[j]
				+ scaledRand() * (d_max[j] - d_min[j]);			
		}
	}
}

//Generate a high quality random number in [0,1]
double DifferentialEvolution::scaledRand() {
	return dist(rng);
}