#include "DynamicsSimulation.h"

template <typename T> int sgn(T val) {
	return (T(0) < val) - (val < T(0));
}

//Constructor
DynamicsSimulation::DynamicsSimulation(double ti, double tf, vector<double> startp, vector<double> endp, double dt, vector<ObstacleInterface*> obs, double s, double mu) {
	//Simulation parameters
	this->ti = ti; //Initial time instant
	this->tf = tf; //Final time instant
	this->dt = dt; //Sampling time interval
	this->s = s; //Future points
	this->tindex = 0; //Time index
	this->mu = mu; //Obstacle threshold

	//Obstacles
	this->obs = obs;

	//Path planning parameters
	this->startp = startp; //Start point
	this->endp = endp; //End point
	
	//States
	this->n =(int)( (tf - ti) / dt + 1); //Number of state vectors
	this->q = vector<vector<double>>(this->n); //Aquired states 
	for (int i = 0; i < this->n; i++) {
		this->q[i] = vector<double>(5); //Five states 
		//q1 (position in x), 
		//q2 (position in y), 
		//q3 (orientation theta),
		//q4 (linear speed v),
		//q5 (angular speed omega),
	}

	//Initial conditions
	this->q[0][0] = startp[0]; //Initial position in x
	this->q[0][1] = startp[1]; //Initial position in y
	this->q[0][2] = startp[2]; //Initial orientation theta
	this->q[0][3] = 0; //Initial linear speed v
	this->q[0][4] = 0; //Initial angular speed omega

	//Control signals
	this->u = vector<vector<double>>(this->n); //Applied control signals
	for (int i = 0; i < (n); i++) {
		this->u[i] = vector<double>(2); //Two control signals
		//u1 (torque of right wheel), 
		//u2 (torque of left wheel), 
	}

	//Control goals
	this->goals = vector<vector<double>>(this->n + n_h); //Control goals (intermediate mobile positions)
	for (int i = 0; i < (n + n_h); i++) {
		this->goals[i] = vector<double>(2); //Two goal components
		//xd (desired x position), 
		//yd (desired y position), 
	}

	//Time
	this->t = vector<double>(this->n); //Time values
	for (int i = 0; i < n; i++) {
		this->t[i] = ti + i * dt; //Given the sampling time
	}

	//Optimization problem parameters
	this->d = 3; //Number of design variables (three: gain v, gain w, sign move)
	this->d_min = {0.0, 0.0, -1.0 };  //Lower bounds for design variables 
	this->d_max = {2.0, 20.0, 1.0};  //Upper bounds for design variables 


	this->p = vector<vector<double>>(this->n); //Calculated design vectors (one per optimization process)
	for (int i = 0; i < n; i++) {
		this->p[i] = vector<double>(d);
	}
}

//Destructor
DynamicsSimulation::~DynamicsSimulation() {
	delete(opt);
}

//Kinematics: output: qdot = [xdot, ydot, thetadot]
vector<double> DynamicsSimulation::differentialMobileDriveKinematics(vector<double> q, double v, double w) {
	//Robot motion simulation
	vector<double> dq = { v * cos(q[2]), v* sin(q[2]), w };
	return dq;
}

//Dynamics: output: qdot = [xdot, ydot, thetadot, vdot, wdot]
vector<double> DynamicsSimulation::differentialMobileDriveDynamics(vector<double> q, double taur, double taul) {
	//Robot motion simulation
	double xdot = q[3] * cos(q[2]);
	double ydot = q[3] * sin(q[2]);
	double thetadot = q[4];
	double vdot = (taur + taul) / (m * r);
	double wdot = L * (taur - taul) / (2.0 * J * r);

	vector<double> dq = { xdot, ydot , thetadot, vdot, wdot};
	return dq;
}

//Simulates the system dynamics, i.e., fills the vector of states q
void DynamicsSimulation::simulateDynamics(int run) {
	clock_t begin = clock(); //Get current SO time
	bool reOptimize = false; //Requires optimization
	int optimization_process = 0; //Count of optimization processes

	//Path features initialization
	collisions = 0;
	arriving_time = 0;
	path_length = 0;
	speed = 0;

	//Bug0 default parameters
	double opt_gv = 0.4; // Optimized v gain
	double opt_gw = 5.0; // Optimized w gain
	double opt_sign = -1.0; // Optimized sign

	//Simualtion variables
	vector<double> qini, qdot; //Initial states and derivated states
	vector<double> op_sol; // Stores the solution of the optimization problem

	//Start simulation
	for (tindex = 0; tindex < n - 1; tindex++) { //For each sampling instant
		qini = q[tindex]; //Get current states

		//Compute distance from robot to obstacles
		vector<double> obs_d = vector<double>();
		for (ObstacleInterface* o : obs) {
			vector<double> o_pos = o->getPosition(t[tindex]);
			double d = sqrt(pow(qini[0] - o_pos[0], 2) + pow(qini[1] - o_pos[1], 2));
			obs_d.push_back(d);

			if (d < 0.15) {
				collisions++; //Collides
			}
		}

		//Search for the nearest obstacle
		double dObst = INFINITY;
		ObstacleInterface* dO = NULL;
		if (obs.size() > 0) {
			dObst = obs_d[0];
			dO = obs[0];
		}
		for (int i = 1; i < obs_d.size(); i++) {
			if (obs_d[i] < dObst) {
				dObst = obs_d[i];
				dO = obs[i];
			}
		}

		//Check nearest obstacle for reoptimization
		if (dObst <= mu) {
			reOptimize = true;
			optimization_process++;
		}

		//Optimization
		if (reOptimize) { // Perform optimization at h and at the first time
			op_sol = opt->startAlgorithm(); //Start metaheuristic optimization

			//Get optimized Bug0 parameters
			opt_gv = op_sol[0];
			opt_gw = op_sol[1];
			opt_sign = op_sol[2] > 0 ? 1.0 : -1.0;

			reOptimize = false;

			p[tindex] = op_sol;
		}
		else {
			p[tindex] = { opt_gv, opt_gw, opt_sign };
		}

		//Control law
		double phiRef, ePhi, dGoal, phiObst;

		vector<double> g;
		if (dObst > mu) { // Drive towards the goal
			phiRef = atan2(endp[1] - qini[1], endp[0] - qini[0]);
			ePhi = phiRef - qini[2];
			dGoal = sqrt(pow(endp[0] - qini[0], 2) + pow(endp[1] - qini[1], 2));
			g = { dGoal / 2.0, 1.0 }; // Control gains
		}
		else {// Drive right arround the obstacle
			vector<double> o_pos = dO->getPosition(t[tindex]);
			phiObst = atan((o_pos[1] - qini[1]) /( o_pos[0] - qini[0]));
			
			phiRef = phiObst + M_PI_2 * opt_sign;
			ePhi = phiRef - qini[2];
			g = { opt_gv, opt_gw };// Control gains
		}

		//Simple control of differential drive
		double v = g[0] * abs(cos(ePhi));
		double w = g[1] * ePhi;

		if (abs(v) > 0.5) {
			v = sgn(v) * 0.5;
		}

		//Compute wheel torques
		double a = 1.0 / (m * r);
		double b = L / (2.0 * J * r);

		double kpv = 50.0;
		double kpw = 50.0;

		double ev = v - qini[3];
		double ew = w - qini[4];

		double v_dot = kpv * ev;
		double w_dot = kpw * ew;

		double taur = v_dot / (2.0 * a) + w_dot / (2.0 * b);
		double taul = v_dot / (2.0 * a) - w_dot / (2.0 * b);
		
		u[tindex][0] = taur;
		u[tindex][1] = taul;

		//Compute dynamic model
		qdot = differentialMobileDriveDynamics(qini, taur, taul);

		//Euler integration
		for (int i = 0; i < 5; i++) {
			q[tindex + 1][i] = qini[i] + qdot[i] * dt;
		}

		//Check if the goal is reached
		double D = sqrt(pow(endp[0] - q[tindex + 1][0], 2) + pow(endp[1] - q[tindex + 1][1], 2));
		if (D < 0.05) {
			cout << "GOAL REACHED!" << endl;
			for (int i = tindex + 1; i < n; i++) {
				for (int j = 0; j < 5; j++) {
					q[i][j] = q[tindex + 1][j];
				}
			}
			break;
		}
	}

	//Calculate path length
	for (int i = 1; i <= tindex; i++) {
		path_length += sqrt(pow(q[i][0] - q[i - 1][0], 2) + pow(q[i][1] - q[i - 1][1], 2));
	}

	//Calculate arriving time
	arriving_time = t[tindex];

	//Calculate speed
	speed = path_length / arriving_time;

	cout << "TOTAL COLISIONS = " << collisions << endl;
	cout << "TOTAL PATH LENGTH = " << path_length << endl;
	cout << "ARRIVING TIME = " << arriving_time << endl;
	cout << "SPEED = " << speed << endl;
	cout << "OPTIMIZATION PROCESS = " << optimization_process << endl;

	clock_t end = clock(); //Get current SO time
	double elapsed_secs = ((double)end - (double)begin) / (double)CLOCKS_PER_SEC; //Compute elapsed simulation time
	execTime = elapsed_secs;
}


//Objective function + constraints (based on kinematics)
vector<double> DynamicsSimulation::evaluate(vector<double> vec) {
	double J = 0; // Objective function and components
	double g = 0; // Constraint violation

	//DESIGN VARIABLES
	double gv = vec[0];
	double gw = vec[1];
	double sign = vec[2] > 0 ? 1.0 : -1.0;

	//DYNAMIC SIMULATION - EULER METHOD
	vector<double> qdot_pp, qini_pp;
	vector<double> q_pp = q[tindex];

	//PREDICTION PARAMETERS
	double dh = dt*s; // Future horizon

	for (int tindex_pp = tindex; tindex_pp < (tindex + s); tindex_pp++) { //For each future sampling instant
		qini_pp = q_pp; //Get current states

		//COMPUTE DISTANCE FROM VEHICLE TO OBSTACLES AND SEARCH FOR THE CLOSEST OBSTACLE
		double cobs_d = INFINITY;
		vector<double> cobs_pos;

		for (ObstacleInterface* o : obs) {
			vector<double> o_pos = o->getPosition(t[tindex_pp]);
			double d = sqrt(pow(qini_pp[0] - o_pos[0], 2) + pow(qini_pp[1] - o_pos[1], 2));

			if (d < cobs_d) {
				cobs_d = d;
				cobs_pos = o_pos;
			}

			if (d < 0.15) { // Detect collisions
				g++;
			}
		}

		//Control law
		double phiRef, ePhi, dGoal, phiObst;

		vector<double> gains;
		if (cobs_d > mu) { // Drive towards the goal
			phiRef = atan2(endp[1] - qini_pp[1], endp[0] - qini_pp[0]);
			ePhi = phiRef - qini_pp[2];
			dGoal = sqrt(pow(endp[0] - qini_pp[0], 2) + pow(endp[1] - qini_pp[1], 2));
			gains = { dGoal / 2.0, 1.0 }; // Control gains
		}
		else {// Drive right arround the obstacle
			phiObst = atan((cobs_pos[1] - qini_pp[1]) / (cobs_pos[0] - qini_pp[0]));
			phiRef = phiObst + M_PI_2*sign;
			ePhi = phiRef - qini_pp[2];
			gains = { gv, gw };
		}

		//Simple control of differential drive
		double v = gains[0] * abs(cos(ePhi));
		double w = gains[1] * ePhi;

		if (abs(v) > 0.5) {
			v = sgn(v) * 0.5;
		}

		//Compute kinematic model
		qdot_pp = differentialMobileDriveKinematics(qini_pp, v, w);

		//Euler integration
		for (int i = 0; i < 5; i++) {
			q_pp[i] = qini_pp[i] + qdot_pp[i] * dt;
		}
	}


	J = sqrt(pow(q_pp[0] - endp[0], 2) + pow(q_pp[1] - endp[1], 2));


	if (J != J) {
		J = 1E200;
	}

	return { J,g };
}

//Optimizer setter
void DynamicsSimulation::setOptimizer(OptimizerInterface *opt) {
	this->opt = opt;
}

//Returns the number of design variables
int DynamicsSimulation::D() {
	return d;
}

//Returns the lower bounds of the design variables
vector<double> DynamicsSimulation::Dmin() {
	return d_min;
}

//Returns the upper bounds of the design variables
vector<double> DynamicsSimulation::Dmax() {
	return d_max;
	
}

