#pragma once

#include <iostream>
#include <fstream>
#include <sstream> 
#include <vector>
#include <string>
#include <time.h>
#include <math.h>
#include <corecrt_math_defines.h>
#include "OptimizationProblemInterface.h"
#include "OptimizerInterface.h"
#include "Obstacle.h"

using namespace std;

class DynamicsSimulation: public OptimizationProblemInterface {
public:
	//CONSTRUCTORS AND DESTRUCTORS
	DynamicsSimulation(double ti, double tf, vector<double> startp, vector<double> endp, double dt, vector<ObstacleInterface*> obs, double s, double mu); //Constructor
	virtual ~DynamicsSimulation(); //Destructor

	//MAIN METHODS
	void simulateDynamics(int run); //Simulates the system dynamics, i.e., fills the vector of states q

	//USEFUL INSTANCE METHODS
	void setOptimizer(OptimizerInterface *opt); //Optimizer setter
											
	//INTERFACE METHODS (OptimizationProblemInterface)
	vector<double> evaluate(vector<double> p); //Objective function + constraints (kinematics version)
	int D(); //Returns the number of design variables
	vector<double> Dmin(); //Returns the lower bounds of the design variables
	vector<double> Dmax(); //Returns the upper bounds of the design variables

	//ROBOT MODELS
	vector<double> differentialMobileDriveKinematics(vector<double> q, double v, double w);
	vector<double> differentialMobileDriveDynamics(vector<double> q, double taur, double taul);

	//PERFORMANCE MEASUREMENTS
	double execTime; //Total simulattion time

private:
	//TIME SIMULATION PARAMETERS
	double ti; //Initial time instant
	double tf; //Final time instant
	double dt; //Sampling time interval
	double s; //Future points
	int tindex; //Time index
	double mu; //Obstacle threshold

	//PATH PLANNING PARAMETERS
	vector<double> startp; //Start point
	vector<double> endp; //End point
	double dtol = 0.05; //Tolerance for the endp reaching (m)

	//PROBLEM PARAMETERS
	int n_opt; //Number of optimization processes
	int n_h; //Number of optimization processes


	//ROBOT PARAMETERS
	double m = 0.75; //Mass (kg)
	double r = 0.024; //Wheel radius (m)
	double L = 0.075; //Car diameter (m)
	double J = 0.001; //Car inertia (kg m ^ 2)

	//STATES
	int n; //Number of states
	vector<vector<double>> q; //Aquired states 

	//CONTROL SIGNALS
	vector<vector<double>> u; //Applied control signals

	//CONTROL GOALS
	vector<vector<double>> goals; //Applied control signals

	//TIME
	vector<double> t; //Time instants

	//OPTIMAL PATH PLANNING PROBLEM
	int d; //Number of design variables
	vector<double> d_min; //Lower bounds for design variables
	vector<double> d_max; //Upper bounds for design variables
	vector<vector<double>> p; //Calculated design vectors

	//OPTIMIZER
	OptimizerInterface *opt; //Optimizer

	//OBSTACLES
	vector<ObstacleInterface*> obs; //Vector of obstacles

public:
	//PATH FEATURES
	int collisions;
	double arriving_time;
	double path_length;
	double speed;
};

