﻿#include "GeneticAlgorithm.h"
#include <math.h>  

//Constructor
GeneticAlgorithm::GeneticAlgorithm(OptimizationProblemInterface* op, int g_max, int np, double pc, double pm, double etac, double etam) : rng(time(0)), dist(0, 1.0)
{
    //Optimization problem
    this->op = op;
    //Optimization problem parameters
    this->d = op->D(); // Problem dimensionality (number of design variables)
    this->d_min = op->Dmin(); // Lower bounds of design variables
    this->d_max = op->Dmax(); // Upper bounds of design variables

    //Algorithm parameters
    this->np = np;//Population size
    this->g_max = g_max;//Maximum generations
    this->pc = pc;//Crossver probability
    this->pm = pm;//Mutation probability
    this->etac = etac;//Distribution index for SBX
    this->etam = etam;//Distribution index for PM

    //Population matrix
    this->x = vector<vector<double>>(np);
    for (int i = 0; i < np; i++) {
        this->x[i] = vector<double>(d+2);
    }
}

GeneticAlgorithm::~GeneticAlgorithm() {
}

//Generate a new population randomly
void GeneticAlgorithm::initializePopulation() {
    for (int i = 0; i < np; i++) {
        for (int j = 0; j < d; j++) {
            x[i][j] = d_min[j]
                + scaledRand() * (d_max[j] - d_min[j]);
        }
    }
}

//Generate a high quality random number in [0,1]
double GeneticAlgorithm::scaledRand() {
    return dist(rng);
}


vector<double> GeneticAlgorithm::startAlgorithm() {
    vector<double> result;
    
    //Randomly initialize the population
    initializePopulation();

    //Evaluate the population
    for (int i = 0; i < np; i++) {
        result = op->evaluate(x[i]);
        x[i][d] = result[0];
        x[i][d + 1] = result[1];
    }

    //Evolutionary cycle
    for (int gen = 0; gen < g_max; gen++)
    {
        crossover();//Parent matching, SBX and PM
        selection();//Sorting and population pruning
    }

    //Return the best chromosome
    vector<double> x_best = x[0];
    x_best.pop_back();
    x_best.pop_back();

    return x_best;
}

void GeneticAlgorithm::crossover()
{
    // Select parents by binary tournament
    vector<int> candidate_indexes_a = vector<int>();
    vector<int> candidate_indexes_b = vector<int>();
    vector<int> parent_indexes = vector<int>();

    //Fill candidate vectors
    for (int i = 0; i < np; i++)
    {
        candidate_indexes_a.push_back(i);
        candidate_indexes_b.push_back(i);
    }

    //Generate a random permutation
    shuffle(begin(candidate_indexes_a), end(candidate_indexes_a), rng);
    shuffle(begin(candidate_indexes_b), end(candidate_indexes_b), rng);

    // Binary comparision of candidates to select parents (use Deb rules)
    for (int i = 0; i < np; i++)
    {

        //If candidates have different constraint violation
        if (x[candidate_indexes_a[i]][d+1] != x[candidate_indexes_b[i]][d+1])

        {
            // Minimum constraint violation is selected
            if (x[candidate_indexes_a[i]][d+1] < x[candidate_indexes_b[i]][d+1])
            {
                parent_indexes.push_back(candidate_indexes_a[i]);
            }
            else
            {
                parent_indexes.push_back(candidate_indexes_b[i]);
            }
        }
        else
        {
            // If candidates have the same constraint violation
            //Minimum objective function is selected
            if (x[candidate_indexes_a[i]][d] < x[candidate_indexes_b[i]][d])
            {
                parent_indexes.push_back(candidate_indexes_a[i]);
            }
            else
            {
                parent_indexes.push_back(candidate_indexes_b[i]);
            }
        }
    }

    int parent_index_1;
    int parent_index_2;

    vector<vector<double>> offsprings;
    vector<double> child_1;
    vector<double> child_2;
    double f_child_1;
    double f_child_2;
    double g_child_1;
    double g_child_2;

    // Generate offpsrings(SBX crossover and polynomial mutation)
    for (int i = 0; i < np / 2; i++)
    {
        // Use crossover probability
        if (scaledRand() < pc)
        {
            // Get the i-th couple
            parent_index_1 = parent_indexes[2 * i];
            parent_index_2 = parent_indexes[2 * i + 1];

            offsprings = sbx(x[parent_index_1], x[parent_index_2]);
            child_1 = offsprings[0];
            child_2 = offsprings[1];

            // Mutate offsprings
            child_1 = poly_mutation(child_1);
            child_2 = poly_mutation(child_2);

            // Evaluate offsprings
            vector<double> result = op->evaluate(child_1);
            child_1[d]=result[0];
            child_1[d+1] =result[1];

            result = op->evaluate(child_2);
            child_2[d] = result[0];
            child_2[d + 1] = result[1];


            //Add offsprings to the population
            x.push_back(child_1);
            x.push_back(child_2);
        }
    }
}

vector<double> GeneticAlgorithm::poly_mutation(vector<double> ind)
{

    double y;
    double xy;
    double yl;
    double yu;
    double delta1;
    double delta2;
    double val;
    double mutPow;
    double rnd;
    double deltaq;


    for (int i = 0; i < d; i++)
    {
        if (scaledRand() <= pm)
        {
            y = ind[i];
            yl = d_min[i];
            yu = d_max[i];
            if (yl == yu)
            {
                y = yl;
            }
            else
            {
                delta1 = (y - yl) / (yu - yl);
                delta2 = (yu - y) / (yu - yl);
                rnd = scaledRand();
                mutPow = 1.0 / (etam + 1.0);
                if (rnd <= 0.5)
                {
                    xy = 1.0 - delta1;
                    val = 2.0 * rnd + (1.0 - 2.0 * rnd) * (pow(xy, etam + 1.0));
                    deltaq = pow(val, mutPow) - 1.0;
                }
                else
                {
                    xy = 1.0 - delta2;
                    val = 2.0 * (1.0 - rnd) + 2.0 * (rnd - 0.5) * (pow(xy, etam + 1.0));
                    deltaq = 1.0 - pow(val, mutPow);
                }
                y = y + deltaq * (yu - yl);

                //Repair variable
                if (y > d_max[i])
                {
                    y = d_min[i] + scaledRand() * (d_max[i] - d_min[i]);
                }
                if (y < d_min[i])
                {
                    y = d_min[i] + scaledRand() * (d_max[i] - d_min[i]);
                }
            }
            ind[i] = y;
        }
    }
    return ind;
}

vector<vector<double>> GeneticAlgorithm::sbx(vector<double> parent1, vector<double> parent2)
{
    double EPS = 1.0e-14;
    vector<vector<double>> offspring = vector<vector<double>>();
    offspring.push_back(parent1);
    offspring.push_back(parent2);

    double valueX1;
    double valueX2;
    double y1;
    double y2;
    double beta;
    double betaq;
    double alpha;
    double c1;
    double c2;
    double ran;
    double lowerBound;
    double upperBound;

    for (int i = 0; i < d; i++)
    {
        valueX1 = parent1[i];
        valueX2 = parent2[i];
        if (scaledRand() <= 0.5)
        {

            if (abs(valueX1 - valueX2) > EPS)
            {
                if (valueX1 < valueX2)
                {
                    y1 = valueX1;
                    y2 = valueX2;
                }
                else
                {
                    y1 = valueX2;
                    y2 = valueX1;
                }

                lowerBound = d_min[i];
                upperBound = d_max[i];

                ran = scaledRand();
                beta = 1.0 + (2.0 * (y1 - lowerBound) / (y2 - y1));
                alpha = 2.0 - pow(beta, -(etac + 1.0));

                if (ran <= (1.0 / alpha))
                {
                    betaq = pow(ran * alpha, (1.0 / (etac + 1.0)));
                }
                else
                {
                    betaq = pow(1.0 / (2.0 - ran * alpha), 1.0 / (etac + 1.0));
                }
                c1 = 0.5 * (y1 + y2 - betaq * (y2 - y1));

                beta = 1.0 + (2.0 * (upperBound - y2) / (y2 - y1));
                alpha = 2.0 - pow(beta, -(etac + 1.0));

                if (ran <= (1.0 / alpha))
                {
                    betaq = pow((ran * alpha), (1.0 / (etac + 1.0)));
                }
                else
                {
                    betaq = pow(1.0 / (2.0 - ran * alpha), 1.0 / (etac + 1.0));
                }
                c2 = 0.5 * (y1 + y2 + betaq * (y2 - y1));

                //Repair variable
                if (c1 > d_max[i])
                {
                    c1 = d_min[i] + scaledRand() * (d_max[i] - d_min[i]);
                }
                if (c1 < d_min[i])
                {
                    c1 = d_min[i] + scaledRand() * (d_max[i] - d_min[i]);
                }

                if (c2 > d_max[i])
                {
                    c2 = d_min[i] + scaledRand() * (d_max[i] - d_min[i]);
                }
                if (c2 < d_min[i])
                {
                    c2 = d_min[i] + scaledRand() * (d_max[i] - d_min[i]);
                }

                if (scaledRand() <= 0.5)
                {
                    offspring[0][i] = c2;
                    offspring[1][i] = c1;
                }
                else
                {
                    offspring[0][i] = c1;
                    offspring[1][i] = c2;
                }
            }
            else
            {
                offspring[0][i] = valueX1;
                offspring[1][i] = valueX2;
            }
        }
        else
        {
            offspring[0][i] = valueX1;
            offspring[1][i] = valueX2;
        }
    }

    return offspring;
}

//Sort the rows in a matrix based on an specific criteria (Deb rules)
void GeneticAlgorithm::sortRows(std::vector<std::vector<double>>& matrix, int d, double ran) {
    std::sort(matrix.begin(),
        matrix.end(),
        [d,ran](const std::vector<double>& lhs, const std::vector<double>& rhs) {
            if (lhs[d + 1] < rhs[d + 1]) {
                return true;
            } else if (lhs[d + 1] > rhs[d + 1]) {
                return false;
            }
            else if (lhs[d + 1]==0 && rhs[d + 1]==0) {
                if (lhs[d] < rhs[d]) {
                    return true;
                }
                else {
                    return false;
                }
            } else if(ran < 0.5) {
                return true;
            }
            else {
                return false;
            }
        });
}

void GeneticAlgorithm::selection()
{
    //Sort population from better to worst (use Deb rules)
    sortRows(x,d, scaledRand());

    //Prune worst chromosomes
    while (x.size() > np)
    {
        x.pop_back();
    }
}
