#include "ParticleSwarmOptimization.h"

//Constructor
ParticleSwarmOptimization::ParticleSwarmOptimization(OptimizationProblemInterface *op, int g_max, int np, double c1, double c2, double w_min, double w_max) : rng(time(0)), dist(0, 1.0) {
	//Optimization problem
	this->op = op;
	//Optimization problem parameters
	this->d = op->D(); //Problem dimensionality (number of design variables)
	this->d_min = op->Dmin(); //Lower bounds of design variables
	this->d_max = op->Dmax(); //Upper bounds of design variables

	//Algorithm parameters
	this->np = np;//Swarm size
	this->g_max = g_max;//Maximum iteration
	this->c1 = c1;//Local knowledge constant
	this->c2 = c2;//Global knowledge constant
	this->w_max = w_max;//Maximum inertia weight
	this->w_min= w_min;//Minimum inertia weight

	//Swarm matrices
	this->x = vector<vector<double>>(np);
	this->x_best = vector<vector<double>>(np);
	this->x_dot = vector<vector<double>>(np);
	for (int i = 0; i < np; i++) {
		this->x[i] = vector<double>(d);
		this->x_dot[i] = vector<double>(np);
		this->x_best[i] = vector<double>(d);
	}
	//Objective function vectors
	this->f_x = vector<double>(np);
	this->f_x_best = vector<double>(np);

	//Constraint vectors
	this->g_x = vector<int>(np);
	this->g_x_best = vector<int>(np);
}

ParticleSwarmOptimization::~ParticleSwarmOptimization() {
}


vector<double> ParticleSwarmOptimization::startAlgorithm() {
	//Randomly initialize the swarm
	initializePopulation();

	//Evaluate the swarm
	vector<double> result;
	for (int i = 0; i < np; i++) {
		result = op->evaluate(x[i]);
		f_x[i] = result[0];
		g_x[i] = result[1];
	}

	//Initialize the personal bests
	for (int i = 0; i < np; i++) {
		x_best[i] = x[i];
		f_x_best[i] = f_x[i];
		g_x_best[i] = g_x[i];
	}

	//Initialize the global best
	x_best_swarm = vector<double>(d);
	f_x_best_swarm = INFINITY;
	g_x_best_swarm = INFINITY;
	initializeSwarmBest();

	double w; //Inertia weight
	bool checkGlobalBest = false;//Flag to determine if it is necessary to check the global best for possible update

	//Evolutionary cycle
	for (int gen = 0; gen < g_max; gen++) {
		//Update the inertia weight	
		w= w_max - ((double)gen / (double)g_max)*(w_max-w_min);

		//Update each particle position using its velocity
		for (int i = 0; i < np; i++) {
			double b1 = scaledRand();
			double b2 = scaledRand();

			//Calculate velocity
			for (int j = 0; j < d; j++) {
				x_dot[i][j] = w*x_dot[i][j] + b1*c1*(x_best[i][j] - x[i][j]) + b2*c2*(x_best_swarm[j] - x[i][j]);
			}

			//Update position
			for (int j = 0; j < d; j++) {
				x[i][j] = x[i][j] + x_dot[i][j];

				//Repair variable
				if (x[i][j] < d_min[j]) {
					//x[i][j] = d_min[j];
					x[i][j]  = d_min[j] + scaledRand() * (d_max[j] - d_min[j]);
				}
				if (x[i][j] > d_max[j]) {
					//x[i][j] = d_max[j];
					x[i][j] = d_min[j] + scaledRand() * (d_max[j] - d_min[j]);
				}
			}

			//Evaluate new position
			result = op->evaluate(x[i]);
			f_x[i] = result[0];
			g_x[i] = result[1];

			//Update the personal best position (use Deb rules)
			if (g_x[i] < g_x_best[i]) {
				x_best[i] = x[i];
				g_x_best[i] = g_x[i];
				f_x_best[i] = f_x[i];
				checkGlobalBest = true;
			}
			else if (g_x[i] > g_x_best[i]) {
				//nop
			}
			else if (g_x[i] == 0 && g_x_best[i] == 0) {
				if (f_x[i] < f_x_best[i]) {
					x_best[i] = x[i];
					g_x_best[i] = g_x[i];
					f_x_best[i] = f_x[i];
					checkGlobalBest = true;
				}
				else {
					//nop
				}
			}
			else if (scaledRand() < 0.5) {
				x_best[i] = x[i];
				g_x_best[i] = g_x[i];
				f_x_best[i] = f_x[i];
				checkGlobalBest = true;
			}
			else {
				//nop
			}

			// If the personal best position was updated
			if (checkGlobalBest) {
				checkGlobalBest = false;
				//Update the global best position if necessary (use Deb rules)
				if (g_x_best[i] < g_x_best_swarm) {
					x_best_swarm = x[i];
					f_x_best_swarm = f_x[i];
					g_x_best_swarm = g_x[i];
				}
				else if (g_x_best[i] > g_x_best_swarm) {
					//nop
				}
				else if (g_x_best[i] == 0 && g_x_best_swarm == 0) {
					if (f_x_best[i] < f_x_best_swarm) {
						x_best_swarm = x[i];
						f_x_best_swarm = f_x[i];
						g_x_best_swarm = g_x[i];
					}
					else {
						//nop
					}
				}
				else if (scaledRand() < 0.5) {
					x_best_swarm = x[i];
					f_x_best_swarm = f_x[i];
					g_x_best_swarm = g_x[i];
				}
				else {
					//nop
				}
			}

		}
	}

	return x_best_swarm; //Return the global best position
}


//Search for the global best position in the whole swarm
void ParticleSwarmOptimization::initializeSwarmBest() {
	for (int i = 0; i < np; i++) {
		if (g_x_best[i] < g_x_best_swarm) {
			x_best_swarm = x[i];
			f_x_best_swarm = f_x[i];
			g_x_best_swarm = g_x[i];
		} 
		else if (g_x_best[i] > g_x_best_swarm) {
			//nop
		}
		else if (g_x_best[i] == 0 && g_x_best_swarm ==0) {
			if (f_x_best[i] < f_x_best_swarm) {
				x_best_swarm = x[i];
				f_x_best_swarm = f_x[i];
				g_x_best_swarm = g_x[i];
			}
			else {
				//nop
			}
		}
		else if (scaledRand() < 0.5) {
			x_best_swarm = x[i];
			f_x_best_swarm = f_x[i];
			g_x_best_swarm = g_x[i];
		}
		else {
			//nop
		}
	}
}

//Randomly initialize the swarm
void ParticleSwarmOptimization::initializePopulation() {
	for (int i = 0; i < np; i++) {
		for (int j = 0; j < d; j++) {
			x[i][j] = d_min[j] + scaledRand() * (d_max[j] - d_min[j]);
			x_dot[i][j] = 0;
		}

	}
}

//Generate a high quality random number in [0,1]
double ParticleSwarmOptimization::scaledRand() {
	return dist(rng);
}